package ibspan.tss.oms;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.TimeZone;
import java.util.zip.CRC32;
import java.util.Set;
import java.util.HashSet;

import com.hp.hpl.jena.ontology.OntDocumentManager;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.ontology.OntModelSpec;
import com.hp.hpl.jena.ontology.FunctionalProperty;
import com.hp.hpl.jena.ontology.InverseFunctionalProperty;
import com.hp.hpl.jena.ontology.Ontology;
import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.query.QueryExecutionFactory;
import com.hp.hpl.jena.query.QueryFactory;
import com.hp.hpl.jena.query.QuerySolution;
import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.rdf.model.ReifiedStatement;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Literal;
import com.hp.hpl.jena.rdf.model.RSIterator;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelMaker;
import com.hp.hpl.jena.shared.Lock;
import com.hp.hpl.jena.vocabulary.RDF;
import com.ibm.adtech.jastor.JastorException;
import com.hp.hpl.jena.util.FileManager;

import ibspan.tss.oms.ontology.ExtInfo;
import ibspan.tss.oms.ontology.OMSFactory;
import ibspan.tss.oms.ontology.OntologySource;
import ibspan.tss.oms.vocabulary.OntologySourceDBVocabulary;
import ibspan.tss.oms.vocabulary.ExtInfoDBVocabulary;
import ibspan.tss.core.semantic.Memory;
import ibspan.tss.core.storage.ModelLoader;
import ibspan.tss.core.storage.ModelLoaderInitException;
import ibspan.tss.domain.ontology.restaurant.Restaurant;
import ibspan.tss.domain.ontology.restaurant.RestaurantFactory;
import ibspan.tss.domain.vocabulary.RestaurantVocabulary;
import ibspan.tss.domain.vocabulary.MoneyVocabulary;
import ibspan.tss.domain.vocabulary.LocationVocabulary;
import ibspan.tss.domain.vocabulary.RestaurantsDBVocabulary;

import org.apache.commons.lang.StringUtils;

public class OMSCore {
	
	private OntModel mRestaurantOnt;
	private OntModel mRestaurantDB;
	private OntModel mExtInfoDB;
	private OntModel mOntSrcDB;
	private Memory mem;
	
	private static final String OMS_CONF_FILE = "file:etc/oms-conf.properties";
	
	private final static int CHECK_UPDATE = 1;
	private final static int KNOWN_UPDATE = 2;
	private final static int REGULAR_UPDATE = 3;
	
	private final int ZERO_ITERATION;
	private final int FINITE_ITERATION;
	
	private final int MIN_SINCE_LAST_UPDATE;
	private final int MIN_NEXT_UPDATE;
	private final int MAX_NEXT_UPDATE;
	
	private final int MIN_INCOMPLETE_CHECK;
	private final int MAX_INCOMPLETE_CHECK;
	
	private final int LEVENSHTEIN_DISTANCE;
	
	public OMSCore() {
		mRestaurantOnt = OMSModel.getInstance().getRestaurantOntModel();
		mRestaurantDB = OMSModel.getInstance().getRestaurantsDBModel();
		mExtInfoDB = OMSModel.getInstance().getExtInfoDBModel();
		mOntSrcDB = OMSModel.getInstance().getOntologySourceDBModel();
		mem = new Memory();
		Properties props = loadProperties(OMS_CONF_FILE);
		ZERO_ITERATION = Integer.valueOf(props.getProperty("omsZeroIteration"));
		FINITE_ITERATION = Integer.valueOf(props.getProperty("omsFiniteIteration"));
		MIN_SINCE_LAST_UPDATE = Integer.valueOf(props.getProperty("omsMinSinceLastUpdate"));
		MIN_NEXT_UPDATE = Integer.valueOf(props.getProperty("omsMinNextUpdate"));
		MAX_NEXT_UPDATE = Integer.valueOf(props.getProperty("omsMaxNextUpdate"));
		MIN_INCOMPLETE_CHECK = Integer.valueOf(props.getProperty("omsMinIncompleteCheck"));
		MAX_INCOMPLETE_CHECK = Integer.valueOf(props.getProperty("omsMaxIncompleteCheck"));
		LEVENSHTEIN_DISTANCE = Integer.valueOf(props.getProperty("omsLevenshteinDistance"));
	}
	
	// Loads properties from the pointed URL.
	private static Properties loadProperties(String confURL) {
		Properties conf = null;
		try {
			java.net.URL url = new java.net.URL(confURL);
			conf = new Properties();
			conf.load(url.openStream());
		} catch (IOException e) {
			throw new RuntimeException("Cannot get properties from configuration file.", e);
		}
		return conf;
	}
	
	public OntModel getModelRestaurantOnt() {
		return this.mRestaurantOnt;
	}
	
	public String getModelRestaurantOntRDF() {
		return mem.model2rdf(this.mRestaurantOnt);
	}
	
	public OntModel rdf2model(String rdf) {
		return mem.rdf2model(rdf);
	}
	
	public String createReifiedStmt(OntModel model, Statement stmt)
	{
		String reifStmt;
    	Resource subject = stmt.getSubject();
    	String prop = stmt.getPredicate().toString();
		RDFNode onode = stmt.getObject();
		String o = onode instanceof Literal ? ((Literal) onode)
				.getString() : ((Resource) onode).getLocalName();
		
        RSIterator iter = model.listReifiedStatements(stmt);
        if (iter.hasNext()) {
        	ReifiedStatement rstmt = iter.nextRS();
        	reifStmt = rstmt.getURI();
        } else {
        	CRC32 crc32 = new CRC32();
	        crc32.update((subject.toString() + prop + o).getBytes());
	        long crc32value = crc32.getValue();
	        reifStmt = ExtInfoDBVocabulary.NS + "ReifStmt_" + crc32value + "_" + System.currentTimeMillis();
	        model.createReifiedStatement(reifStmt, stmt);
        }
        return reifStmt;
	}
	
	public String createReifiedStmt(OntModel model, Statement stmt, String reifStmtURI)
	{
		String reifStmt;
    	Resource subject = stmt.getSubject();
    	String prop = stmt.getPredicate().toString();
		RDFNode onode = stmt.getObject();
		String o = onode instanceof Literal ? ((Literal) onode)
				.getString() : ((Resource) onode).getLocalName();
		
        RSIterator iter = model.listReifiedStatements(stmt);
        if (iter.hasNext()) {
        	ReifiedStatement rstmt = iter.nextRS();
        	reifStmt = rstmt.getURI();
	        System.out.println("The statement: " + subject.toString() + " -> " + prop.toString() + " -> " + o +
	        		" already existed as: " + reifStmt);
        } else {
	        reifStmt = reifStmtURI;
	        model.createReifiedStatement(reifStmt, stmt);
	        System.out.println("The statement: " + subject.toString() + " -> " + prop.toString() + " -> " + o +
	        		" was reified as: " + reifStmt);
        }
        return reifStmt;
	}
	
	public void addPrimaryExtInfoToReifStmt(String reifStmt, String hasURL, RDFNode node) {
		Resource rOntSrc;
		Resource rExtInfo;
		try {
			rOntSrc = mOntSrcDB.createResource(OntologySourceDBVocabulary.NS + changeURLtoLocalName(hasURL));
			if (!mOntSrcDB.getBaseModel().contains(rOntSrc, RDF.type, OntologySource.TYPE)) {
				mOntSrcDB.begin();
				OntologySource ontSrc = OMSFactory.createOntologySource(rOntSrc, mOntSrcDB);
				ontSrc.setHasURL(hasURL);
				ontSrc.setHasCF(0.0f);
				ontSrc.setHasCorrectTripples(1);
				ontSrc.setHasWrongTripples(0);
				mOntSrcDB.commit();
			}
			else {
				mOntSrcDB.begin();
				OntologySource ontSrc = OMSFactory.getOntologySource(rOntSrc, mOntSrcDB);
				ontSrc.setHasCorrectTripples(ontSrc.getHasCorrectTripples()+1);
				ontSrc.setHasCF(computeHasCFOntSrc(ontSrc.getHasCorrectTripples(),ontSrc.getHasWrongTripples()));
				mOntSrcDB.commit();
			}
			rExtInfo = mExtInfoDB.createResource(reifStmt);
			if (!mExtInfoDB.getBaseModel().contains(rExtInfo, RDF.type, ExtInfo.TYPE)) {
				mExtInfoDB.begin();
				ExtInfo extInfo = OMSFactory.createExtInfo(rExtInfo, mExtInfoDB);
				extInfo.setHasUpdateDateTime(
					getUpdateCheckConf(CHECK_UPDATE,MIN_SINCE_LAST_UPDATE,
						MIN_NEXT_UPDATE,ZERO_ITERATION));
				extInfo.setHasCF(0.0f);
				extInfo.setHasMainOntSrc(rOntSrc);
				if (node.equals(Restaurant.TYPE)) {
					extInfo.setHasIncompleteCheckDateTime(getIncompleteCheckConf(MIN_INCOMPLETE_CHECK));
				}
				mExtInfoDB.commit();
			} else {
				mExtInfoDB.begin();
				ExtInfo extInfo = OMSFactory.createExtInfo(rExtInfo, mExtInfoDB);
				extInfo.addHasPositiveOntSrc(rOntSrc);
				mExtInfoDB.commit();
			}
		} catch (JastorException e) {
			mExtInfoDB.abort();
			mOntSrcDB.abort();
			throw new RuntimeException("Cannot add extInfo to the reified statement.", e);
		}	
		
	}
	
	private String changeURLtoLocalName(String hasURL) {
		hasURL = hasURL.replace("/", "0");
		hasURL = hasURL.replace(":", "10");
		hasURL = hasURL.replace("#", "110");
		hasURL = hasURL.replace(".", "1110");
		hasURL = hasURL.replace("&", "1111");
		return hasURL;
	}
	
	private ReifiedStatement retrievalReifStmt(String reifStmt) { 
		ReifiedStatement rs = null;
		Resource r = mRestaurantDB.getResource(reifStmt);
		if (r.canAs(ReifiedStatement.class))
		{
			rs = (ReifiedStatement) r.as(ReifiedStatement.class);
		}
	    return rs;
	}
	
	public void deleteReifiedStmt(OntModel model, Statement stmt)
	{
		model.removeAllReifications(stmt);	
	}
	
	private float computeHasCFOntSrc(int correct, int wrong) {
		return (correct-wrong)/(correct+wrong);
	}
	
	private float computeHasCFStmt(int positive, int negative) {
		return (positive+1-negative)/(positive+1+negative);
	}
	
	private String getUpdateCheckConf(int typeUpdate,double sinceLastUpdate, 
			double nextUpdate, int iterUpdate) {
		String updatePrimaryConf = "";
		Date now = new Date();
		Calendar c1 = Calendar.getInstance();
		c1.setTime(now);
		for (int i=1;i<=7;i++)
		{
			c1.add(Calendar.DAY_OF_YEAR, (int)sinceLastUpdate);
			updatePrimaryConf += dateFormat().format(c1.getTime()) + ";" + typeUpdate +
			"," + convToIntIfPoss(sinceLastUpdate) + 
			"," + convToIntIfPoss(nextUpdate) + 
			"," + iterUpdate + "&";
		}
		return updatePrimaryConf;
	}
	
	private String getIncompleteCheckConf(double nextUpdate) {
		String conf = "";
		Date now = new Date();
		Calendar c1 = Calendar.getInstance();
		c1.setTime(now);
		c1.add(Calendar.DAY_OF_YEAR, (int)nextUpdate);
		conf += dateFormat().format(c1.getTime()) + ";" + convToIntIfPoss(nextUpdate*2) + "&";
		return conf;
	}
	
	private String findDateTimeToCheck(String datesConf) {
		Date now = new Date();
		String rtnDatesConf = "";
		String rtnDateForUpdate = null;
		Date maxDateForUpadte = null;
		boolean firstRun = true;
		Date d = null;
		String[] dates = datesConf.split("&");
		for (int i=0;i<dates.length;i++)
		{
			String[] datePar = dates[i].split(";");
			try {
				d = dateFormat().parse(datePar[0]);
			} catch (ParseException e) {
				e.printStackTrace();
			}
			if (now.after(d)) {
				if (firstRun) {
					firstRun = false;
					rtnDateForUpdate = dates[i];
					maxDateForUpadte = d;
				} else if (!firstRun && d.after(maxDateForUpadte)) {
					rtnDateForUpdate = dates[i];
					maxDateForUpadte = d;
				}
			} else {
				rtnDatesConf += dates[i] + "&";
			}
		}
		
		return rtnDateForUpdate + "&&" + rtnDatesConf;
	}
	
	private Date addToDateTime(Date date, double val) {
		int addVal = (int)(24*val);
		double diff;
		Calendar c1 = Calendar.getInstance();
		c1.setTime(date);
		c1.add(Calendar.HOUR_OF_DAY, addVal);
		val = 24*val;
		diff = val-addVal;
		if (diff>0) {
			val = 60*diff;
			addVal = (int)(val);
			c1.add(Calendar.MINUTE, addVal);
			diff = val-addVal;
			if (diff>0) {
				addVal = (int)(60*diff);
				c1.add(Calendar.SECOND, addVal);
			}
		}
		return c1.getTime();
	}
	
	
	private Map<Resource, Map<Resource, String>> getResourcesFromDB(Property prop, int cntTripplesForUpdate, int checkType) {
		
		Map<Resource, Map<Resource, String>> mapOntSrc = new HashMap<Resource, Map<Resource, String>>();
		Map<Resource, String> mapRes;
		ExtInfo extInfo;
		int maxResults = cntTripplesForUpdate;

		// create query string
		String queryString = "SELECT ?x ?y ?z " +
	    		"WHERE " +
	    		"{ ?x <" + prop + "> ?y . " +
	    		" ?x <" + ExtInfo.hasMainOntSrcProperty + "> ?z . }"; 

	    Query query = QueryFactory.create(queryString);
	    QueryExecution qexec = QueryExecutionFactory.create(query, mExtInfoDB);
	    try {
	    	ResultSet results = qexec.execSelect();
			int i = 0;
			while (results.hasNext() && i < maxResults) {	
	    		QuerySolution soln = results.nextSolution();
	    		
	    		Literal lDate = soln.getLiteral("y");
	    		if (lDate.getString().length()>0)
	    		{
	        		String[] resultUpdate = findDateTimeToCheck(lDate.getString()).split("&&");
	        		if (!resultUpdate[0].equals("null")) {
	        			i++;
	        			Resource r = soln.getResource("x");
	        			Resource rOntSrc = soln.getResource("z");
	        			
	        			// set new UpdatePirmary value without dateTime to check
	        			try {
	        				extInfo = OMSFactory.getExtInfo(r, mExtInfoDB);
	        				if (resultUpdate.length>1 && !resultUpdate[1].equals("null")) {
	        					if (checkType==0)
	        						extInfo.setHasUpdateDateTime(resultUpdate[1]);
	        					else
	        						extInfo.setHasIncompleteCheckDateTime(resultUpdate[1]);
	        				} else {
	        					if (checkType==0)
	        						extInfo.setHasUpdateDateTime("");
	        					else
	        						extInfo.setHasIncompleteCheckDateTime("");
	        				}
	        			} catch (JastorException e) {
	        				throw new RuntimeException("Cannot get ExtInfo resource.", e);
	        			}	
	        			
	        			// put resources for giving OntSource  
	        			mapRes = (Map<Resource, String>) mapOntSrc.get(rOntSrc);
	        			if (mapRes == null) {
	        				mapRes = new HashMap<Resource,String>();
	        				mapRes.put(r, resultUpdate[0]);
	        				mapOntSrc.put(rOntSrc, mapRes);
	        			} else {
	        				mapRes.put(r, resultUpdate[0]);
	        			}
	        		}
	    		}
	    	}
	    } finally {qexec.close();}

        return mapOntSrc;
	}
	
	public Map<String, Map<Long, String>> createModelsForUpdate(int cntTripples) {
		
		Map<Resource, Map<Resource, String>> mapOntSrc = getResourcesFromDB(ExtInfo.hasUpdateDateTimeProperty, cntTripples, 0);
		Map<String, Map<Long, String>> mapModels = new HashMap<String, Map<Long, String>>();
		Map<Long, String> mapOntology;
		Map mapNSRes = mRestaurantDB.getNsPrefixMap();
		Map mapNSExtInfo = mExtInfoDB.getNsPrefixMap();
		OntologySource ontologySource = null;
		
		for (Map.Entry<Resource, Map<Resource, String>> entry : mapOntSrc.entrySet()) {
        	OntModel model = ModelFactory.createOntologyModel(OntModelSpec.OWL_MEM_RULE_INF);
        	model.setNsPrefixes(mapNSExtInfo);
        	model.setNsPrefixes(mapNSRes);
            for (Map.Entry<Resource, String> entry2 : entry.getValue().entrySet()) {
            	try {
	            	ReifiedStatement rs = retrievalReifStmt(entry2.getKey().getURI());	    		
	    			model.add(rs.getStatement().getSubject(), RDF.type, Restaurant.TYPE);
		    		model.add(rs.getStatement());
		    		model.createReifiedStatement(rs.getURI(), rs.getStatement());
	    			ExtInfo extInfo = OMSFactory.createExtInfo(entry2.getKey(), model);
	    			extInfo.setHasUpdateDateTime(entry2.getValue());
	    		} catch (JastorException e) {
	    			throw new RuntimeException("Cannot create or get data from OMSFactory.", e);
	    		}
            }
        	Long currentTimeM = System.currentTimeMillis();
        	mapOntology = new HashMap<Long, String>();
			mapOntology.put(currentTimeM, mem.model2rdf(model));
			try {
				ontologySource = OMSFactory.getOntologySource(entry.getKey(), mOntSrcDB);
				mapModels.put(ontologySource.getHasURL(), mapOntology);
    		} catch (JastorException e) {
    			throw new RuntimeException("Cannot put data to mapModels Map.", e);
    		}		
        }
		return mapModels;
	}
	
	public Map<String, Map<Long, String>> createIncompleteModels(int cntTripples) {
		
		Map<Resource, Map<Resource, String>> mapOntSrc = getResourcesFromDB(ExtInfo.hasIncompleteCheckDateTimeProperty, cntTripples, 1);
		Map<String, Map<Long, String>> mapModels = new HashMap<String, Map<Long, String>>();
		Map<Long, String> mapOntology;
		Map mapNSRes = mRestaurantDB.getNsPrefixMap();
		Map mapNSExtInfo = mExtInfoDB.getNsPrefixMap();
		OntologySource ontologySource = null;
		
		for (Map.Entry<Resource, Map<Resource, String>> entry : mapOntSrc.entrySet()) {
        	OntModel model = ModelFactory.createOntologyModel(OntModelSpec.OWL_MEM_RULE_INF);
        	model.setNsPrefixes(mapNSExtInfo);
        	model.setNsPrefixes(mapNSRes);
            for (Map.Entry<Resource, String> entry2 : entry.getValue().entrySet()) {
            	try {
	            	ReifiedStatement rs = retrievalReifStmt(entry2.getKey().getURI());	    		
	    			model.add(rs.getStatement().getSubject(), RDF.type, Restaurant.TYPE);
		    		model.add(rs.getStatement());
		    		model.createReifiedStatement(rs.getURI(), rs.getStatement());
	    			ExtInfo extInfo = OMSFactory.createExtInfo(entry2.getKey(), model);
	    			extInfo.setHasIncompleteCheckDateTime(entry2.getValue());	
	    	        for (Iterator i = mRestaurantOnt.listObjectProperties(); i.hasNext(); ) {
	    	        	Property prop = (Property) i.next();
	    				Restaurant r = RestaurantFactory.getRestaurant(
	    						rs.getStatement().getSubject(), mRestaurantDB);
	    				StmtIterator it = r.resource().listProperties(prop); 
	    				if (!it.hasNext()) {
	    					model.add(rs.getStatement().getSubject(), prop, "");
	    				}
	    	        }
	    	        for (Iterator i = mRestaurantOnt.listDatatypeProperties(); i.hasNext(); ) {
	    	        	Property prop = (Property) i.next();
	    				Restaurant r = RestaurantFactory.getRestaurant(
	    						rs.getStatement().getSubject(), mRestaurantDB);
	    				StmtIterator it = r.resource().listProperties(prop); 
	    				if (!it.hasNext()) {
	    					model.add(rs.getStatement().getSubject(), prop, "");
	    				}
	    	        }
	    			
	    		} catch (JastorException e) {
	    			throw new RuntimeException("Cannot create or get data from OMSFactory.", e);
	    		}
            }
        	Long currentTimeM = System.currentTimeMillis();
        	mapOntology = new HashMap<Long, String>();
			mapOntology.put(currentTimeM, mem.model2rdf(model));
			try {
				ontologySource = OMSFactory.getOntologySource(entry.getKey(), mOntSrcDB);
				mapModels.put(ontologySource.getHasURL(), mapOntology);
    		} catch (JastorException e) {
    			throw new RuntimeException("Cannot put data to mapModels Map.", e);
    		}		
        }
		return mapModels;
	}
	
	private int compareFunctionalProp(OntModel mRequest, OntModel mResponse, 
			Set<Property> listFunctionalProp) {
		int cntUpdates = 0;
		try {
			for (Iterator<Restaurant> it = RestaurantFactory.getAllRestaurant(
					mRequest).iterator(); it.hasNext();) {
				Restaurant rReq = it.next();
				for (StmtIterator itStmt = rReq.resource().listProperties(); itStmt.hasNext();) {
					Statement stmtReq = itStmt.nextStatement();
					// property is functional or inversefunctional
		            if (listFunctionalProp.contains(stmtReq.getPredicate())) {
						RDFNode onodeReq = stmtReq.getObject();
						String valReq = onodeReq instanceof Literal ? ((Literal) onodeReq)
							.getString() : ((Resource) onodeReq).getLocalName();
						Restaurant rResp = RestaurantFactory.getRestaurant(
							rReq.resource().getURI(), mResponse);
						RSIterator iterRS = mRequest.listReifiedStatements(stmtReq);
						ReifiedStatement rs = iterRS.nextRS();
		            	boolean bEquals = true;
		            	StmtIterator itStmtResp = null;
		            	try {
		            		itStmtResp = rResp.resource().listProperties(stmtReq.getPredicate());
		            	} catch (Exception e) {
							deleteStmtEverywhere(stmtReq, rs);
							break;
		            	}
		            	Statement stmtResp = itStmtResp.nextStatement();
						RDFNode onodeResp = stmtResp.getObject();
						String valResp = onodeResp instanceof Literal ? ((Literal) onodeResp)
								.getString() : ((Resource) onodeResp).getLocalName();
						if (!valReq.equals(valResp))
							bEquals = false;
						
						ExtInfo extInfo = OMSFactory.getExtInfo(rs.getURI(), mRequest);
						String[] tUpdateDT = extInfo.getHasUpdateDateTime().split(";");
						Date updateDateTime = null;
						try {
							updateDateTime = dateFormat().parse(tUpdateDT[0]);
						} catch (ParseException e) {
							e.printStackTrace();
						}
						String[] tPar = tUpdateDT[1].split(",");
						Integer iTypeUpdate = Integer.parseInt(tPar[0]);
						Double dSinceLastUpdate = Double.parseDouble(tPar[1]);
						Double dNextUpdate = Double.parseDouble(tPar[2]);
						Integer iIterUpdate = Integer.parseInt(tPar[3]);
						if (checkConsistencyOntReasoner(rReq.resource().getURI(), stmtResp)) {
							cntUpdates += updateMethod(updateDateTime, iTypeUpdate, dSinceLastUpdate, 
									dNextUpdate, iIterUpdate, bEquals,
									stmtReq, stmtResp, rs);
						}
		            }
				}
			}
		} catch (JastorException e) {
			throw new RuntimeException("Cannot compare functional properties.", e);
		}
		
		return cntUpdates;
	}
	
	private int compareNotFunctionalProp(OntModel mRequest, OntModel mResponse, 
			Set<Property> listFunctionalProp) {
		int cntUpdates = 0;
		Map<Property, Boolean> mapProp;
		try {
			for (Iterator<Restaurant> it = RestaurantFactory.getAllRestaurant(
					mRequest).iterator(); it.hasNext();) {
				Restaurant rReq = it.next();
				mapProp = new HashMap<Property, Boolean>();
				// first we have to check if given property was changed
				for (StmtIterator itStmt = rReq.resource().listProperties(); itStmt.hasNext();) {
					Statement stmtReq = itStmt.nextStatement();
					// property is not functional or not inversefunctional
		            if (!listFunctionalProp.contains(stmtReq.getPredicate()) && !stmtReq.getPredicate().equals(RDF.type)) {
		            	RDFNode onodeReq = stmtReq.getObject();
		            	String valReq = onodeReq instanceof Literal ? ((Literal) onodeReq)
							.getString() : ((Resource) onodeReq).getLocalName();
						Restaurant rResp = RestaurantFactory.getRestaurant(
								rReq.resource().getURI(), mResponse);
		            	if (!mapProp.containsKey(stmtReq.getPredicate()))
		            		mapProp.put(stmtReq.getPredicate(), true);
		            	boolean bExists = false;
						for (StmtIterator itStmtResp = rResp.resource().listProperties(stmtReq.getPredicate()); 
									itStmtResp.hasNext();) {
							Statement stmtResp = itStmtResp.nextStatement();
							RDFNode onodeResp = stmtResp.getObject();
							String valResp = onodeResp instanceof Literal ? ((Literal) onodeResp)
									.getString() : ((Resource) onodeResp).getLocalName();						
							if (valReq.equals(valResp)) {
								bExists = true;
								break;
							}
						}
						if (!bExists) {
							mapProp.remove(stmtReq.getPredicate());
							mapProp.put(stmtReq.getPredicate(), false);
						}
		            }
				}
				
				// now we make upadte with properly boolean value for given property
				for (StmtIterator itStmt = rReq.resource().listProperties(); itStmt.hasNext();) {
					Statement stmtReq = itStmt.nextStatement();
					// property is not functional or not inversefunctional
		            if (!listFunctionalProp.contains(stmtReq.getPredicate()) && !stmtReq.getPredicate().equals(RDF.type)) {
		            	RDFNode onodeReq = stmtReq.getObject();
		            	String valReq = onodeReq instanceof Literal ? ((Literal) onodeReq)
							.getString() : ((Resource) onodeReq).getLocalName();
						RSIterator iterRS = mRequest.listReifiedStatements(stmtReq);
						ReifiedStatement rs = iterRS.nextRS();
						ExtInfo extInfo = OMSFactory.getExtInfo(rs.getURI(), mRequest);
						String[] tUpdateDT = extInfo.getHasUpdateDateTime().split(";");
						Date updateDateTime = null;
						try {
							updateDateTime = dateFormat().parse(tUpdateDT[0]);
						} catch (ParseException e) {
							e.printStackTrace();
						}
						String[] tPar = tUpdateDT[1].split(",");
						Integer iTypeUpdate = Integer.parseInt(tPar[0]);
						Double dSinceLastUpdate = Double.parseDouble(tPar[1]);
						Double dNextUpdate = Double.parseDouble(tPar[2]);
						Integer iIterUpdate = Integer.parseInt(tPar[3]);
						Restaurant rResp = RestaurantFactory.getRestaurant(
								rReq.resource().getURI(), mResponse);
						Statement stmtResp = null;
		            	boolean bExists = false;
						for (StmtIterator itStmtResp = rResp.resource().listProperties(stmtReq.getPredicate()); 
									itStmtResp.hasNext();) {
							Statement stmtRespTemp = itStmtResp.nextStatement();
							RDFNode onodeResp = stmtRespTemp.getObject();
							String valResp = onodeResp instanceof Literal ? ((Literal) onodeResp)
									.getString() : ((Resource) onodeResp).getLocalName();						
							if (valReq.equals(valResp)) {
								stmtResp = stmtRespTemp;
								bExists = true;
								break;
							}
						}
						if (!bExists) {
							deleteStmtEverywhere(stmtReq, rs);
						} else {
							if (checkConsistencyOntReasoner(rReq.resource().getURI(), stmtResp)) {
								cntUpdates += updateMethod(updateDateTime, iTypeUpdate, dSinceLastUpdate, 
									dNextUpdate, iIterUpdate, mapProp.get(stmtReq.getPredicate()),
									stmtReq, stmtResp, rs);
							}
						}
		            }
				}
				
			}
		} catch (JastorException e) {
			throw new RuntimeException("Cannot compare not functional properties.", e);
		}
		return cntUpdates;
	}
	
	private int updateMethod(Date updateDateTime, Integer iTypeUpdate, Double dSinceLastUpdate, 
			Double dNextUpdate, Integer iIterUpdate, boolean bEquals,
			Statement stmtReq, Statement stmtResp, ReifiedStatement rs) {
		
		int cntUpdates = 0;
		String newUpdateConf = "";
		try {
			if (iTypeUpdate==CHECK_UPDATE) {
				if (iIterUpdate==ZERO_ITERATION && bEquals) {
					ExtInfo extInfo2 = OMSFactory.getExtInfo(rs.getURI(), mExtInfoDB);
					if (extInfo2.getHasUpdateDateTime().equals("") && dNextUpdate>=MAX_NEXT_UPDATE) {
						newUpdateConf = getNewUpdateConf(
							dateFormat().format(addToDateTime(updateDateTime, dNextUpdate)),
							REGULAR_UPDATE,MAX_NEXT_UPDATE,MAX_NEXT_UPDATE,ZERO_ITERATION);
						addUpdateConf(rs, newUpdateConf);
					} 
					else if (extInfo2.getHasUpdateDateTime().equals("")) {
						newUpdateConf = getUpdateCheckConf(
							CHECK_UPDATE, dSinceLastUpdate*2.0, dNextUpdate*2.0, ZERO_ITERATION);
						addUpdateConf(rs, newUpdateConf);
					}
				}
				else if (iIterUpdate<FINITE_ITERATION && !bEquals) {
					newUpdateConf = getNewUpdateConf(
						dateFormat().format(addToDateTime(updateDateTime,
							(dNextUpdate - dSinceLastUpdate/2.0))),
						CHECK_UPDATE,dSinceLastUpdate/2.0,
						dNextUpdate,++iIterUpdate);
					makeUpdate(stmtReq, stmtResp, rs, newUpdateConf);
					cntUpdates++;
				} else if (iIterUpdate>ZERO_ITERATION && iIterUpdate<FINITE_ITERATION && bEquals) {
					newUpdateConf = getNewUpdateConf(
						dateFormat().format(addToDateTime(updateDateTime,
							dSinceLastUpdate/2.0)),
						CHECK_UPDATE,dSinceLastUpdate/2.0,
						dNextUpdate,++iIterUpdate);
					addUpdateConf(rs, newUpdateConf);
				} else if (iIterUpdate==FINITE_ITERATION && !bEquals) {
					newUpdateConf = getNewUpdateConf(
						dateFormat().format(addToDateTime(updateDateTime,dNextUpdate)),
						KNOWN_UPDATE,dSinceLastUpdate*Math.pow(2.0, iIterUpdate),
						dNextUpdate,ZERO_ITERATION);
					makeUpdate(stmtReq, stmtResp, rs, newUpdateConf);
					cntUpdates++;
				} else if (iIterUpdate==FINITE_ITERATION && bEquals) {
					newUpdateConf = getNewUpdateConf(
						dateFormat().format(addToDateTime(updateDateTime,dSinceLastUpdate)),
						KNOWN_UPDATE,dSinceLastUpdate*Math.pow(2.0, iIterUpdate),
						dNextUpdate,ZERO_ITERATION);
					addUpdateConf(rs, newUpdateConf);
				}
			} else if (iTypeUpdate==KNOWN_UPDATE) {
				if (bEquals) {
					newUpdateConf = getNewUpdateConf(
						dateFormat().format(addToDateTime(updateDateTime,dSinceLastUpdate/2.0)),
						CHECK_UPDATE,dSinceLastUpdate,
						dNextUpdate,ZERO_ITERATION);
					addUpdateConf(rs, newUpdateConf);
				} else {
					newUpdateConf = getNewUpdateConf(
						dateFormat().format(addToDateTime(updateDateTime,dNextUpdate)),
						KNOWN_UPDATE,dSinceLastUpdate,
						dNextUpdate,iIterUpdate);
					makeUpdate(stmtReq, stmtResp, rs, newUpdateConf);
					cntUpdates++;
				}
			} else if (iTypeUpdate==REGULAR_UPDATE) {
				if (bEquals) {
					newUpdateConf = getNewUpdateConf(
						dateFormat().format(addToDateTime(updateDateTime, dNextUpdate)),
						REGULAR_UPDATE,dSinceLastUpdate,dNextUpdate,iIterUpdate);
					addUpdateConf(rs, newUpdateConf);
				} else {
					newUpdateConf = getUpdateCheckConf(
						CHECK_UPDATE, MIN_SINCE_LAST_UPDATE,
						MIN_NEXT_UPDATE, ZERO_ITERATION);
					makeUpdate(stmtReq, stmtResp, rs, newUpdateConf);
					cntUpdates++;
				}
			}
		} catch (JastorException e) {
			throw new RuntimeException("Cannot compare properties", e);
		}
		return cntUpdates;
	}
	
	public int compareRequestResponseModels(OntModel mRequest, OntModel mResponse) {
	
		int cntUpdates = 0;

		Set<Property> listFunctionalProp = getFunctionalProp();
		
        cntUpdates = compareFunctionalProp(mRequest, mResponse, listFunctionalProp);
        cntUpdates += compareNotFunctionalProp(mRequest, mResponse, listFunctionalProp);
        
		return cntUpdates;
	}
	
	public int insertIncompleteData(String ontSrcURL, OntModel mResponse) {
		
		int cntInsert = 0;
		boolean incompleteData;
		try {
			for (Iterator<Restaurant> it = RestaurantFactory.getAllRestaurant(
					mResponse).iterator(); it.hasNext();) {
				Restaurant rRes = it.next();
				incompleteData = false;
				for (StmtIterator itStmt = rRes.resource().listProperties(); itStmt.hasNext();) {
					Statement stmt = itStmt.nextStatement();
					RDFNode onode = stmt.getObject();
					String o = onode instanceof Literal ? ((Literal) onode)
							.getString() : ((Resource) onode).getLocalName();
					if (o.length()>0) {
						if (checkConsistencyOntReasoner(rRes.resource().getURI(), stmt)) {
							mRestaurantDB.add(stmt);
					    	String reifS = createReifiedStmt(mRestaurantDB, stmt);
					    	if (reifS!=null)
					    		addPrimaryExtInfoToReifStmt(reifS, ontSrcURL, stmt.getObject());
					    	cntInsert++;
					    	incompleteData = true;
						}
					}
				}
				StmtIterator itStmt = rRes.resource().listProperties(RDF.type);
				Statement stmt = itStmt.nextStatement();
				RSIterator iterRS = mResponse.listReifiedStatements(stmt);
				ReifiedStatement rs = iterRS.nextRS();
				ExtInfo extInfo = OMSFactory.getExtInfo(rs.getURI(), mResponse);
				String[] confDT = extInfo.getHasIncompleteCheckDateTime().split(";");
				Date dateTime = null;
				try {
					dateTime = dateFormat().parse(confDT[0]);
				} catch (ParseException e) {
					e.printStackTrace();
				}
				Double dNextCheck = Double.parseDouble(confDT[1]);
				if (incompleteData) {
					dNextCheck = (((dNextCheck/2.0)<MIN_INCOMPLETE_CHECK) ? MIN_INCOMPLETE_CHECK : dNextCheck/2.0);
					addIncompleteConf(rs, getNewIncompleteConf(dateFormat().format
							(addToDateTime(dateTime, dNextCheck)), dNextCheck));
				} else {
					addIncompleteConf(rs, getNewIncompleteConf(dateFormat().format
							(addToDateTime(dateTime, dNextCheck)), 
							(((dNextCheck*2.0)>MAX_INCOMPLETE_CHECK) ? MAX_INCOMPLETE_CHECK : dNextCheck*2.0)));
				}
				
			}
		} catch (JastorException e) {
			throw new RuntimeException("Cannot insert incomplete data.", e);
		}
        
		return cntInsert;
	}
	
	private Set<Property> getFunctionalProp() {
		Set<Property> listFunctionalProp = new HashSet<Property>();
        for (Iterator i = mRestaurantOnt.listFunctionalProperties(); i.hasNext(); ) {
        	FunctionalProperty f = (FunctionalProperty) i.next();
        	listFunctionalProp.add(f);
        }
        for (Iterator i = mRestaurantOnt.listInverseFunctionalProperties(); i.hasNext(); ) {
        	InverseFunctionalProperty f = (InverseFunctionalProperty) i.next();
        	listFunctionalProp.add(f);
        }
        return listFunctionalProp;
	}
	
	private String getNewUpdateConf(String dateTime, int typeUpdate,double sinceLastUpdate, 
			double nextUpdate, int iterUpdate) {
		return dateTime + ";" + typeUpdate + "," + convToIntIfPoss(sinceLastUpdate) + "," +
			convToIntIfPoss(nextUpdate) + "," + iterUpdate + "&";
	}
	
	private String getNewIncompleteConf(String dateTime, double nextUpdate) {
		return dateTime + ";" + convToIntIfPoss(nextUpdate) + "&";
	}
	
	private String convToIntIfPoss(double val) {
		String fmt = "#####0";
		DecimalFormat df = new DecimalFormat(fmt);
		return (((val-(int)val)==0)?df.format(val):val).toString();
	}
	
	private DateFormat dateFormat() {
		DateFormat dfm = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
		dfm.setTimeZone(TimeZone.getTimeZone("Europe/Zurich"));
		return dfm;
	}
	
	private void deleteStmtEverywhere(Statement stmt, ReifiedStatement rs) {
		// delete reified statement
		deleteReifiedStmt(mRestaurantDB, stmt);
		// remove statement
		mRestaurantDB.remove(stmt);
		try {
			// get ExtInfo
			ExtInfo extInfo = OMSFactory.getExtInfo(rs.getURI(), mExtInfoDB);
			// delete ExtInfo
			extInfo.removeStatements();
		} catch (JastorException e) {
			throw new RuntimeException("Cannot delete extInfo.", e);
		}
	}
	
	private void makeUpdate(Statement stmt_old, Statement stmt_new, 
			ReifiedStatement rs, String newUpdateConf) {
		
		String reifStmt;
		
		// delete reified statement
		deleteReifiedStmt(mRestaurantDB, stmt_old);

		// remove old statement and add new one
		mRestaurantDB.remove(stmt_old);
		mRestaurantDB.add(stmt_new);
		
		// create reified statement
		reifStmt = createReifiedStmt(mRestaurantDB, stmt_new);
		
		// update reified statement for given stmt_new
		try {
			// get old ExtInfo
			ExtInfo extInfo = OMSFactory.getExtInfo(rs.getURI(), mExtInfoDB);
			
			// create new ExtInfo
			Resource rExtInfo = mExtInfoDB.createResource(reifStmt);			
			mExtInfoDB.begin();
			ExtInfo extInfoNew = OMSFactory.createExtInfo(rExtInfo, mExtInfoDB);
			extInfoNew.setHasUpdateDateTime(extInfo.getHasUpdateDateTime() + newUpdateConf);
			extInfoNew.setHasCF(extInfo.getHasCF());
			extInfoNew.setHasMainOntSrc(extInfo.getHasMainOntSrc());
			mExtInfoDB.commit();

			// delete old ExtInfo
			extInfo.removeStatements();
			
		} catch (JastorException e) {
			throw new RuntimeException("Cannot set new extInfo.", e);
		}
	}
	
	private void addUpdateConf(ReifiedStatement rs, String newUpdateConf) {
		// update reified statement for given stmt_new
		try {
			ExtInfo extInfo = OMSFactory.getExtInfo(rs.getURI(), mExtInfoDB);
			extInfo.setHasUpdateDateTime(extInfo.getHasUpdateDateTime() + newUpdateConf);
		} catch (JastorException e) {
			throw new RuntimeException("Cannot set extInfo.", e);
		}
	}
	
	private void addIncompleteConf(ReifiedStatement rs, String newConf) {
		try {
			ExtInfo extInfo = OMSFactory.getExtInfo(rs.getURI(), mExtInfoDB);
			extInfo.setHasIncompleteCheckDateTime(newConf);
		} catch (JastorException e) {
			throw new RuntimeException("Cannot set extInfo.", e);
		}
	}
	
	public int newDataIndexing(Map<String, String> mapNewModels) {
		int cntNewStmt = 0;
		Set<Property> listFunctionalProp = getFunctionalProp();
		for (Map.Entry<String, String> entry : mapNewModels.entrySet()) {
			System.out.println(entry.getKey().toString());
			try {
				for (Iterator<Restaurant> it = RestaurantFactory.getAllRestaurant(
						mem.rdf2model(entry.getValue())).iterator(); it.hasNext();) {
					Restaurant r = it.next();
					Restaurant restDB = RestaurantFactory.getRestaurant(
							r.resource().getURI(), mRestaurantDB);
					if (restDB==null) {
						for (StmtIterator itStmt = r.resource().listProperties(); itStmt.hasNext();)
						{
							Statement stmt = itStmt.nextStatement();
							if (checkConsistencyOntReasoner(r.resource().getURI(), stmt)) {
								mRestaurantDB.add(stmt);
						    	String reifS = createReifiedStmt(mRestaurantDB, stmt);
						    	if (reifS!=null)
						    		addPrimaryExtInfoToReifStmt(reifS, entry.getKey().toString(), stmt.getObject());
						    	cntNewStmt++;
							}
						}
					}
					else 
					{
						for (StmtIterator itStmt = r.resource().listProperties(); itStmt.hasNext();)
						{
							Statement stmt = itStmt.nextStatement();
						    Resource subject = stmt.getSubject();
						    Property p = stmt.getPredicate();
							StmtIterator itStmtDB = restDB.resource().listProperties(p); 
							if (!itStmtDB.hasNext())
							{
								if (checkConsistencyOntReasoner(r.resource().getURI(), stmt)) {
									mRestaurantDB.add(stmt);
							    	String reifS = createReifiedStmt(mRestaurantDB, stmt);
							    	if (reifS!=null)
							    		addPrimaryExtInfoToReifStmt(reifS, entry.getKey().toString(), stmt.getObject());
							    	cntNewStmt++;
								}
							}
							else 
							{
								RDFNode onode = stmt.getObject();
						        String val = onode instanceof Literal ? ((Literal) onode)
							        	.getString() : ((Resource) onode).getLocalName();
								Statement stmtDB = itStmtDB.nextStatement();
								RDFNode onodeDB = stmtDB.getObject();
						        String valDB = onodeDB instanceof Literal ? ((Literal) onodeDB)
							        	.getString() : ((Resource) onodeDB).getLocalName();
							    if ((StringUtils.getLevenshteinDistance(val, valDB)<=LEVENSHTEIN_DISTANCE
							    		&& listFunctionalProp.contains(p))
							    		|| val.equals(valDB)) {
							    	System.out.println("Statements\n" + val + "\n" + valDB + "\nare equal");
							    } else {
							    	String reifS1 = createReifiedStmt(mRestaurantDB, stmtDB);
									ExtInfo extInfo1 = OMSFactory.getExtInfo( reifS1, mExtInfoDB);
									OntologySource ontSrc1 = OMSFactory.getOntologySource(entry.getKey().toString(), mOntSrcDB);
									float prod1 = ontSrc1.getHasCF() * extInfo1.getHasCF();
									float prod2 = 0.0f;
									if (prod1>prod2) {
										ontSrc1.setHasWrongTripples(ontSrc1.getHasWrongTripples()+1);
										extInfo1.addHasNegativeOntSrc(ontSrc1);
									} else {
										deleteReifiedStmt(mExtInfoDB, stmtDB);
										mRestaurantDB.remove(stmtDB);
										if (checkConsistencyOntReasoner(r.resource().getURI(), stmt)) {
											mRestaurantDB.add(stmt);
											String reifS = createReifiedStmt(mRestaurantDB, stmt);
											if (reifS!=null)
												addPrimaryExtInfoToReifStmt(reifS, entry.getKey().toString(), stmt.getObject());
											cntNewStmt++;
										}
									}
							    }
							}
					    }	
					}
				}
			} catch (JastorException e) {
				throw new RuntimeException("Exception while retrieving restaurants from model ", e);
			}
		}
		return cntNewStmt;
	}
	
	public boolean checkConsistencyOntReasoner(String uri, Statement stmtToCheck)
	{
		OMSReasoner reasoner = new OMSReasoner();
		Boolean consistent = false;
		
		mRestaurantDB.enterCriticalSection(Lock.READ);
		try {
			OntModel m = ModelFactory.createOntologyModel();
			Restaurant r = RestaurantFactory.getRestaurant(
					RestaurantsDBVocabulary.NS + uri, mRestaurantDB);	
			for (StmtIterator it2 = r.resource().listProperties(); it2.hasNext();)
			{
			    Statement stmt = it2.nextStatement();
			    Resource subject = stmt.getSubject();
			    Property p = stmt.getPredicate();
				RDFNode onode = stmt.getObject();
				m.add(subject, p, onode);
			}
			m.setNsPrefixes(mRestaurantDB.getNsPrefixMap());
			Ontology ont = m.createOntology(null);
			ont.addImport( m.createResource( "http://www.ibspan.waw.pl/tss/Restaurant" ) );
			reasoner.setModel(m);
			reasoner.addStatement(stmtToCheck);
			consistent = reasoner.checkConsistency();
			System.out.println("Consistent=" + consistent.toString());
		} catch (JastorException e) {
			throw new RuntimeException("Exception while retrieving restaurants from db: ", e);
		} finally {
			mRestaurantDB.leaveCriticalSection();
		}
		return consistent;
	}
}
